#!/usr/bin/env python

#
# Copyright 2015, Olivier MATZ <zer0@droids-corp.org>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the University of California, Berkeley nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import email.header
import re

# pylint: disable=deprecated-module
# see https://www.logilab.org/ticket/2481
import string

class VarFormatter(string.Formatter):
    """
    Simple formatter that does not throw exception if the
    variable does not exist. In this case, it is replaced by an
    empty string.
    """
    def __init__(self):
        string.Formatter.__init__(self)

    def get_field(self, field_name, args, kwargs):
        try:
            return string.Formatter.get_field(self, field_name, args, kwargs)
        except (KeyError, AttributeError):
            return None, field_name

    def format_field(self, value, spec):
        if value is None:
            return ''
        return string.Formatter.format_field(self, value, spec)

def headers_to_unicode(headers):
    """
    Convert mail headers into a unicode dictionary

    :arg email.message.Message headers:
      The email headers
    """
    unicode_headers = {}
    for key, hdr in headers.items():
        try:
            value, encoding = email.header.decode_header(hdr)[0]
        except email.header.HeaderParseError:
            try:
                # try to workaround badly formatted RFC2047 tokens
                hdr = re.sub(r"(==)(?!$)", u"= =", hdr)
                value, encoding = email.header.decode_header(hdr)[0]
            except email.header.HeaderParseError:
                # fallback to wrong decoding
                value, encoding = hdr, 'utf-8'
        if encoding is None:
            value = unicode(value)
        else:
            value = value.decode(encoding, errors="replace")
        unicode_headers[key] = value
    return unicode_headers

def highest_fetch_level(fetch_list):
    """
    Return the highest fetch level for a mail.
    """
    if "all" in fetch_list:
        return "all"
    if "part1" in fetch_list:
        return "part1"
    if "headers" in fetch_list:
        return "headers"
    return "no"
